# -*- perl -*-

#
# $Id: BBBikeDraw.pm,v 1.2 1998/08/01 21:02:58 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 1998 Slaven Rezic. All rights reserved.
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

package BBBikeDraw;
use strict;
use Strassen;

my $gd_version;
my($grey_bg, $white, $yellow, $red, $green, $darkgreen, $darkblue, $lightblue);
my %color;

sub new {
    my($pkg, %args) = @_;

    my $self = {};
    $self->{Fh}        = delete $args{Fh} || \*STDOUT;
    $self->{Geometry}  = delete $args{Geometry};
    $self->{Coords}    = delete $args{Coords};
    $self->{Draw}      = delete $args{Draw};
    $self->{Startname} = delete $args{Startname};
    $self->{Zielname}  = delete $args{Zielname};
    $self->{Strassen}  = delete $args{Strassen};

    bless $self, $pkg;

    $self->init;
}

sub new_from_cgi {
    my($pkg, $q, $fh, $str) = @_;
    new($pkg,
	Fh        => $fh,
	Geometry  => $q->param('geometry'),
	Coords    => [ split(/;/, $q->param('coords')) ],
	Draw      => [ $q->param('draw') ],
	Startname => $q->param('startname'),
	Zielname  => $q->param('zielname'),
	Strassen  => $str,
       );
}

sub init {
    my $self = shift;

    my($geometry) = $self->{Geometry};

    eval q{
	local $SIG{'__DIE__'};
	require GD;
    };
    return undef if ($@);

    eval q{local $SIG{'__DIE__'};
	   $gd_version = $GD::VERSION;
       };

    my($w, $h) = (640, 480);
    if (defined $geometry) {
	($w, $h) = split(/x/, $geometry);
    }
    my $im = GD::Image->new($w,$h);

    $grey_bg   = $im->colorAllocate(153,153,153);
    $white     = $im->colorAllocate(255,255,255);
    $yellow    = $im->colorAllocate(255,255,0);
    $red       = $im->colorAllocate(255,0,0);
    $green     = $im->colorAllocate(0,255,0);
    $darkgreen = $im->colorAllocate(0,128,0);
    $darkblue  = $im->colorAllocate(0,0,128);
    $lightblue = $im->colorAllocate(186,213,247);

    %color = (H  => $yellow,
	      N  => $white,
	      NN => $green,
	      S  => $darkgreen,
	      SA => $darkgreen,
	      SB => $darkgreen,
	      SC => $darkgreen,
	      U  => $darkblue,
	      UA => $darkblue,
	      UB => $darkblue,
	      W  => $lightblue,
	      F  => $white,
	      P  => $green, # Konflikt mit NN
	     );

    $im->interlaced('true');

    $self->{Image}  = $im;
    $self->{Width}  = $w;
    $self->{Height} = $h;

    $self;
}

sub dimension_from_route {
    my $self = shift;
    my(@coords) = @{ $self->{Coords} };
    my $im = $self->{Image};
    my $fh = $self->{Fh};
    my @c1;
    my($min_x, $min_y, $max_x, $max_y);
    foreach (@coords) {
	my($x, $y) = split(/,/, $_);
	if (!defined $min_x || $x < $min_x) {
	    $min_x = $x;
	}
	if (!defined $max_x || $x > $max_x) {
	    $max_x = $x;
	}
	if (!defined $min_y || $y < $min_y) {
	    $min_y = $y;
	}
	if (!defined $max_y || $y > $max_y) {
	    $max_y = $y;
	}
	push @c1, [$x,$y];
    }

    if ((!defined $max_x && !defined $min_x) ||
	$max_x == $min_x || $max_y == $min_y) {
	$im->string(GD->gdLargeFont, 10, 10, "Empty image!", $white);
	print $fh $im->gif;
	die "Empty image";
    }

    # etwas Luft lassen
    $min_x -= int(0.06*($max_x-$min_x));
    $min_y -= int(0.06*($max_y-$min_y));
    $max_x += int(0.06*($max_x-$min_x));
    $max_y += int(0.06*($max_y-$min_y));

    $self->{Min_x} = $min_x;
    $self->{Min_y} = $min_y;
    $self->{Max_x} = $max_x;
    $self->{Max_y} = $max_y;
    $self->{C1}    = \@c1;
}

sub set_dimension {
    my $self = shift;
    ($self->{Min_x}, $self->{Max_x}, $self->{Min_y}, $self->{Max_y}) = @_;
}

sub create_transpose {
    my $self = shift;
    my($w, $h) = ($self->{Width}, $self->{Height});
    my($min_x, $min_y, $max_x, $max_y) =
      ($self->{Min_x}, $self->{Min_y}, $self->{Max_x}, $self->{Max_y});
    my($xk, $yk) = ($w/($max_x-$min_x), $h/($max_y-$min_y));
    my $aspect = ($max_x-$min_x)/($max_y-$min_y);
    my($delta_x, $delta_y) = ($min_x, $min_y);
    if ($aspect < $w/$h) {
	$xk *= $aspect/($w/$h);
	$delta_x -= ($w/$xk-$max_x+$min_x)/2;
    } else {
	$yk /= $aspect/($w/$h);
	$delta_y -= ($h/$yk-$max_y+$min_y)/2;
    }
    my $transpose = sub {
	my($x, $y) = @_;
	(($x-$delta_x)*$xk, $h-($y-$delta_y)*$yk);
    };
    $self->{Transpose} = $transpose;
}

sub draw_map {
    my $self = shift;
    my $im = $self->{Image};
    my $transpose = $self->{Transpose};
    # Netze zeichnen
    my @netz;
    my(%str_draw);
    my $title_draw;
    foreach (@{$self->{Draw}}) {
	if ($_ eq 'title' &&
	    $self->{Startname} ne '' &&
	    $self->{Zielname}  ne '') {
	    $title_draw = 1;
	} else {
	    $str_draw{$_} = 1;
	}
    }
    # Reihenfolge (von unten nach oben): Gewsser, Straen, U-, S-Bahn
    push @netz, new Strassen 'wasserstrassen' if $str_draw{'wasser'};
    my $str = $self->{Strassen};
    if (!defined $str) {
	$str = new Strassen "strassen";
    }
    push @netz, $str if $str_draw{'str'};
    foreach (
	     ['ubahn',           'ubahn'],
	     ['sbahn',           'sbahn'],
	    ) {
	push @netz, new Strassen $_->[0] if $str_draw{$_->[1]}
    }

    foreach my $strecke (@netz) {
	$strecke->init;
	while(1) {
	    my $s = $strecke->next_obj;
	    last if $s->is_empty;
	    if ($s->category =~ /^F:(.*)/) {
		if ($1 eq 'I') {
		    next; # Inseln vorerst ignorieren
		}
		my $c = $color{$1} || 'white';
		my $poly = GD::Polygon->new();
		for(my $i = 0; $i <= $#{$s->coords}; $i++) {
		    $poly->addPt(&$transpose(@{$s->coord_as_list($i)}));
		}
		$im->filledPolygon($poly, $c);
	    } else {
		for(my $i = 0; $i < $#{$s->coords}; $i++) {
		    my($x1, $y1, $x2, $y2) = (@{$s->coord_as_list($i)},
					      @{$s->coord_as_list($i+1)});
		    $im->line(&$transpose($x1, $y1),
			      &$transpose($x2, $y2),
			      $color{$s->category} || $white,
			     );
		}
	    }
	}
    }

    my($xw1, $yw1) = &$transpose(0, 0);
    my($xw2, $yw2) = &$transpose(40, 40);
    my($xw, $yw) = ($xw2-$xw1, $yw2-$yw1);

    foreach my $points (['ubahn', 'ubahnhof'],
			['sbahn', 'sbahnhof'],
		       ) {
	if ($str_draw{$points->[0]}) {
	    my $p = new Strassen $points->[1];
	    my $type = ($points->[0] eq 'ubahn' ? 'u' : 's');
	    $p->init;
	    while(1) {
		my $s = $p->next_obj;
		last if $s->is_empty;
		if ($type eq 'u') {
		    my($x0, $y0) = @{$s->coord_as_list(0)};
		    my($x1, $y1) = &$transpose($x0-20, $y0-20);
		    my($x2, $y2) = &$transpose($x0+20, $y0+20);
		    # Achtung: y2 und y1 vertauschen!
		    $im->filledRectangle($x1, $y2, $x2, $y1, $darkblue);
		} else {
		    my($x, $y) = &$transpose(@{$s->coord_as_list(0)});
		    $im->arc($x, $y, $xw, $yw, 0, 360, $darkgreen);
		}
	    }
	}
    }

    $self->{TitleDraw} = $title_draw;
}

sub draw_route {
    my $self = shift;
    my $im        = $self->{Image};
    my $transpose = $self->{Transpose};
    my(@c1)       = @{ $self->{C1} };

    # Route
    for(my $i = 0; $i < $#c1; $i++) {
	my($x1, $y1, $x2, $y2) = (@{$c1[$i]}, @{$c1[$i+1]});
	$im->line(&$transpose($x1, $y1),
		  &$transpose($x2, $y2), $red);
    }

    if ($self->{TitleDraw}) {
	my $s = patch_string($self->{Startname} . " -> "
			     . $self->{'Zielname'});
	for ([-1, 0], [1, 0], [0, 1], [0, -1]) {
	    $im->string(GD->gdMediumBoldFont, 1+$_->[0], 1+$_->[1],
			$s, $darkblue);
	}
	$im->string(GD->gdMediumBoldFont, 1, 1,
		    $s, $white);
    }
}

sub make_imagemap {
    my $self = shift;
    my $fh = shift || die "No file handle supplied";

    my $transpose = $self->{Transpose};
    my $str = $self->{Strassen};
    if (!defined $str) {
	$str = new Strassen "strassen";
    }

    require URI::Escape;

    print $fh <<EOF;
<script language=javascript>
<!--
function s(text) {
  self.status=text;
  return true;
}
// -->
</script>
EOF
    print $fh "<map name=\"map\">";
    $str->init;
    while(1) {
	my $s = $str->next_obj;
	last if $s->is_empty;
	if ($s->category !~ /^F/) {
	    my(@polygon1, @polygon2);
	    my($dx, $dy, $c);
	    my($x1, $y1, $x2, $y2);
	    for(my $i = 0; $i < $#{$s->coords}; $i++) {
		($x1, $y1, $x2, $y2) = 
		  (&$transpose(@{$s->coord_as_list($i)}),
		   &$transpose(@{$s->coord_as_list($i+1)}));
		$dx = $x2-$x1;
		$dy = $y2-$y1;
		$c = CORE::sqrt($dx*$dx + $dy*$dy)/2;
		if ($c == 0) { $c = 0.00001; }
		$dx /= $c;
		$dy /= $c;
		push    @polygon1, int($x1-$dy), int($y1+$dx);
		unshift @polygon2, int($x1+$dy), int($y1-$dx);
	    }
	    # letzter Punkt
	    push    @polygon1, int($x2-$dy), int($y2+$dx);
	    unshift @polygon2, int($x2+$dy), int($y2-$dx);

	    print $fh
	      "<area href=\"" . URI::Escape::uri_escape($s->name)
		. "\" ",
		"shape=poly ",
		"coords=\"" . join(",", @polygon1, @polygon2,
				   $polygon1[0], $polygon1[1]) . "\" ",
		"onMouseOver=\"return s('" . $s->name . "');\" ",
		">\n";
	}
    }
    print $fh "</map>";
}

sub flush {
    my $self = shift;
    my $fh = $self->{Fh};
    print $fh $self->{Image}->gif;
}

sub patch_string {
    if (defined $gd_version and $gd_version >= 1.16) {
	shift;
    } else {
	$_ = shift;
	s//ae/g;
	s//oe/g;
	s//ue/g;
	s//Ae/g;
	s//Oe/g;
	s//Ue/g;
	s//ss/g;
	s//e/g;
	$_;
    }
}

1;
