# -*- perl -*-

#
# $Id: Karte.pm,v 1.3 1998/05/26 11:26:31 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 1998 Slaven Rezic. All rights reserved.
# This package is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

package Karte;
#use AutoLoader 'AUTOLOAD';

use strict;
use vars qw(%map @map $cache_root);

$cache_root = $ENV{TMPDIR} || '/tmp';

sub preload {
    my(@types) = @_;
    if ($types[0] eq ':all') {
	@types = qw(Standard
		    Berlinmap1996 Berlinmap1997 Berlinmap1998
		    Satmap SatmapGIF GISmap
		   );
    }
    my $karte;
    foreach $karte (@types) {
	my $module = "Karte::$karte";
	eval "require $module";
	if ($@) {
	    warn $@;
	} else {
	    my $o = eval "\$" . $module . "::obj";
	    next if exists $map{$o->token};
	    push @map, $o->token;
	    $map{$o->token} = $o;
	}
    }
}

sub name     { shift->{Name} }
sub token    { shift->{Token} }
sub mimetype { shift->{Mimetype} }
sub coordsys { shift->{Coordsys} }

sub fs_dir    { shift->{Fs_dir} }
sub cache_dir { shift->{Cache_dir} }
sub root_url  { shift->{Root_URL} }

sub x0 { shift->{X0} }
sub x1 { shift->{X1} }
sub x2 { shift->{X2} }
sub y0 { shift->{Y0} }
sub y1 { shift->{Y1} }
sub y2 { shift->{Y2} }

sub width  { shift->{Width} }
sub height { shift->{Height} }

sub scrollregion { @{shift->{Scrollregion}} }

# gibt an, ob die Karte mit Umgebung gezeichnet werden kann
sub noenvironment { shift->{NoEnvironment} }

sub to_ppm {
    my $self = shift;
    ($self->mimetype eq 'image/gif'
     ? 'giftoppm'
     : ($self->mimetype eq 'image/jpeg'
	? 'djpeg'
	: die "Unknown to_ppm for " . $self->mimetype
       )
    );
}

sub ext {
    my $self = shift;
    ($self->mimetype eq 'image/gif'
     ? 'gif'
     : ($self->mimetype eq 'image/jpeg'
	? 'jpg'
	: die "Unknown extension for " . $self->mimetype
       )
    );
}

# Erhhen/Erniedrigen der Kartenkoordinaten
sub incx {
    my($self, $c, $inc) = @_;
    $c + $inc; 
}

sub incy {
    my($self, $c, $inc) = @_;
    $c + $inc; 
}


# Transformationen von und zum normalen Koordinatensystem
sub map2standard {
    my($self, $oldx, $oldy) = @_;
    (int($self->x0 + $oldx*$self->x1 + $oldy*$self->x2),
     int($self->y0 + $oldx*$self->y1 + $oldy*$self->y2)
    );
}

sub standard2map {
    my($self, $newx, $newy) = @_;
    (int((($newx-$self->x0)*$self->y2-($newy-$self->y0)*$self->x2)/
	 ($self->x1*$self->y2-$self->y1*$self->x2)),
     int((($newx-$self->x0)*$self->y1-($newy-$self->y0)*$self->x1)/
	 ($self->x2*$self->y1-$self->x1*$self->y2))
    );
}

# Transformation zwischen beliebigen Koordinatensystemen
sub map2map {
    my($from, $to, $oldx, $oldy) = @_;
    $to->standard2map($from->map2standard($oldx, $oldy));
}

sub scale_coeff {
    my $self = shift;
    my($x0, $y0) = $self->map2standard(0, 0);
    my($x1, $y1) = $self->map2standard(1, 0);
    sqrt(($x1-$x0)*($x1-$x0) + ($y1-$y0)*($y1-$y0));
}

# statische Funktion
sub coordsys2obj {
    my($coordsys) = @_;
    my($k, $v);
    while(($k,$v) = each %Karte::map) {
	my $match_coordsys = $v->coordsys;
	return $v if (defined $match_coordsys &&
		      $match_coordsys eq $coordsys);
    }
    undef;
}

1;
