# -*- perl -*-

#
# $Id: PLZ.pm,v 1.8 1998/08/01 21:07:49 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 1998 Slaven Rezic. All rights reserved.
# This package is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

package PLZ;
use strict;
use vars qw(@plzfile $OLD_AGREP);
use locale;

@plzfile = 
  ((map { ("$_/Berlin.small.data", "$_/data/Berlin.small.data") } @INC),
   (map { ("$_/Berlin.data", "$_/data/Berlin.data") } @INC),
   "/usr/www/soc/plz/Berlin.data",
   "/home/pub/lib/plz/Berlin.data",
  );
$OLD_AGREP = 0 unless defined $OLD_AGREP;

sub new {
    my($class, $file) = @_;
    my $self = {};
    if (!defined $file) {
	foreach (@plzfile) {
	    if (-r $_) {
		$file = $_;
		if (!open(DATA, $file)) {
		    $file = "$file.gz";
		    next if !-f $file || !open(DATA, "zcat $file|");
		    $self->{IsGzip} = 1;
		} else {
		    $self->{IsGzip} = 0;
		}
		my($line) = <DATA>;
		my(@l) = split(/\|/, $line);
		if (@l > 4) {
		    $self->{DataFmt} = 'normal';
		} else {
		    $self->{DataFmt} = 'reduced';
		}
		close DATA;
		last;
	    }
	}
    }
    return undef if !defined $file || !-r $file;
    $self->{File} = $file;
    bless $self, $class;
}

sub open {
    my($self, %args) = @_;
    my $file = $args{File} || $self->{File};
    if ($file ne $self->{Data}) {
	my @data;
	open(PLZ, $file)
	  or die "Die Datei $file kann nicht geffnet werden: $!";

	my $code = <<'EOF';
	while(<PLZ>) {
	    chomp;
	    my(@l) = split(/\|/, $_);
EOF
	my $push_code;
	if ($self->{DataFmt} eq 'reduced') {
	    $push_code = q{push @data, [@l[0 .. 2]]};
	} else {
	    $push_code = q{push @data, [@l[0, 1, 4]]};
	}
	$code .= $push_code . <<'EOF';
	}
EOF
        eval $code;
	close PLZ;
	$self->{Data} = \@data;
	$self->{File} = $file;
	undef $self->{NameHash};
	undef $self->{PlzHash};
    }
}

sub make_plz_re {
    my($self, $plz) = @_;
    if ($self->{DataFmt} eq 'reduced') {
	'^[^|]*|[^|]*|' . $plz;
    } else {
	'^[^|]*|[^|]*|[^|]*|[^|]*|' . $plz . '|';
    }
}

# XXX Dudenstr. wird nur als Kreuzberg, nicht auch als Tempelhof ausgegeben
# XXX make gzip-aware
# Ausgabe: Array von Referenzen [strasse, bezirk, plz]
sub look {
    my($self, $str, %args) = @_;
    my $file = $args{File} || $self->{File};
    my @res;

    my $grep_type = ($args{Agrep} ? 'agrep' : 'grep');
    my $plz_field = ($self->{DataFmt} eq 'reduced' ? 2 : 4);
    if ($grep_type eq 'agrep') {
	if ($OLD_AGREP ||
	    (!$args{Noextern} && !_is_in_path('agrep'))) {
	    $args{Noextern} = 1;
	}
	if ($args{Noextern}) {
	    eval q{local $SIG{'__DIE__'};
		   require String::Approx;
		   String::Approx->VERSION(2.7);
	       };
	    if ($@) {
		$grep_type = 'grep';
	    }
	}
    }
    if ($grep_type eq 'grep') {
	if (!$args{Noextern} && !_is_in_path('grep')) {
	    $args{Noextern} = 1;
	}
    }
    if (!$args{Noextern}) {
	unless ($args{Noquote}) {
	    if ($grep_type eq 'grep') {
		$str =~ s/([\\.*\[\]])/\\$1/g; # quote metacharacters
	    } else {
		$str =~ s/(.)/\\$1/g;
	    }
	    $str = "^$str";
	}
	my(@grep_args) = ('-i', $str, $file);
	if ($grep_type eq 'agrep' && $args{Agrep}) {
	    unshift @grep_args, "-$args{Agrep}";
	}
	CORE::open(PLZ, "-|") or exec $grep_type, @grep_args;
	my %res;
	while(<PLZ>) {
	    chomp;
	    my(@to_push) = (split(/\|/, $_))[0,1,$plz_field];
	    if (!exists $res{$to_push[0]}->{$to_push[1]} &&
		!exists $res{$to_push[0]}->{$to_push[2]}) {
		push(@res, [@to_push]);
		last if defined $args{Max} and $args{Max} < $#res;
		$res{$to_push[0]}->{$to_push[1]}++;
		$res{$to_push[0]}->{$to_push[2]}++;
	    }
	}
	close PLZ;
    } else {
	CORE::open(PLZ, $file)
	  or die "Die Datei $file kann nicht geffnet werden: $!";
	if ($args{Agrep}) {
	    chomp(my @data = <PLZ>);
	    close PLZ;
	    my %res;
	    foreach (String::Approx::amatch($str,
					    ['i', $args{Agrep}], @data)) {
		my(@to_push) = (split(/\|/, $_))[0,1,$plz_field];
		if (!exists $res{$to_push[0]}->{$to_push[1]} &&
		    !exists $res{$to_push[0]}->{$to_push[2]}) {
		    push(@res, [@to_push]);
		    last if defined $args{Max} and $args{Max} < $#res;
		    $res{$to_push[0]}->{$to_push[1]}++;
		    $res{$to_push[0]}->{$to_push[2]}++;
		}
	    }
	} else {
	    $str = "^$str" unless $args{Noquote};
	    $str =~ s/\|/\\|/g;
	    my %res;
	    while(<PLZ>) {
		chomp;
		my(@to_push) = (split(/\|/, $_))[0,1,$plz_field];
		if (/$str/i &&
		    !exists $res{$to_push[0]}->{$to_push[1]} &&
		    !exists $res{$to_push[0]}->{$to_push[2]}) {
		    push(@res, [@to_push]);
		    last if defined $args{Max} and $args{Max} < $#res;
		    $res{$to_push[0]}->{$to_push[1]}++;
		    $res{$to_push[0]}->{$to_push[2]}++;
		}
	    }
	    close PLZ;
	}
    }
    @res;
}

# Versucht zunchst ohne Agrep, zu matchen und dann inkrementell bis
# $args{Agrep}.
# Wenn "strasse" ausgeschrieben wurde, wird es noch einmal mit "str."
# versucht.
sub look_loop {
    my($self, $str, %args) = @_;
    my $max_agrep = delete $args{Agrep} || 0;
    my @matchref;
    my $loop;
    for $loop (0 .. 1) {
	@matchref = $self->look($str, %args);
	if (@matchref == 0) {
	    my $agrep = 1;
	    while (!@matchref && $agrep <= $max_agrep) {
		@matchref = $self->look($str, %args, Agrep => $agrep);
		$agrep++;
	    }
	}
	if (@matchref == 0 && $str =~ /stra(ss|)e/i) {
	    $str =~ s/stra(ss|)e/str./;
	} else {
	    last;
	}
    }
    @matchref;
}

sub make_name_hash {
    my $self = shift;
    my %hash;
    my $i = 0;
    foreach (@{$self->{Data}}) {
	$hash{$_->[0]} = $i;
	$i++;
    }
    $self->{NameHash} = \%hash;
}

sub make_plz_hash {
    my $self = shift;
    my %hash;
    my $i = 0;
    foreach (@{$self->{Data}}) {
	$hash{$_->[2]} = $i;
	$i++;
    }
    $self->{PlzHash} = \%hash;
}

sub _is_in_path {
    my($prog) = @_;
    if ($ENV{PATH}) {
	foreach (split(/:/, $ENV{PATH})) {
	    return $_ if -x "$_/$prog";
	}
    }
    undef;
}

1;

