# -*- perl -*-

#
# $Id: Telefonbuch.pm,v 1.5 1998/07/03 14:03:37 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 1998 Slaven Rezic. All rights reserved.
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

# experimentell
package Telefonbuch;

use BBBikeUtil;
use Karte::GIS;
use strict;
use vars qw($VERBOSE @datadirs $cdromdir $maxrec);

use constant pagesize => 0x2000;

@datadirs = qw(/cdrom/database /cd/database /CDROM/database);
if (defined $cdromdir) {
    unshift @datadirs, "$cdromdir/database";
}
if ($^O =~ /mswin32/i) {
    push @datadirs, qw(D:/database E:/database F:/database);
}

my $str_base = "bstrd";
my $tel_base = "btp1d";

my %old_linear;

#print join("\n", map { join(", ", @$_) } search_string($search)), "\n";

sub exists {
    foreach (@datadirs) {
	return $_ if -d $_;
    }
    undef;
}

sub check_exists {
    my($top) = @_;
    if (!Telefonbuch::exists()) {
	require Tk::Dialog;
	$top->Dialog
	  (-title => 'Fehler',
	   -text  => 'Die Telefonbuch-CD-ROM ist anscheinend nicht eingelegt',
	   -bitmap => 'error')->Show;
	return undef;
    } else {
	1;
    }
}

sub tk_str_dialog {
    my($top, $mark_sub) = @_;
    return if !check_exists($top);
    require Tk::LabEntry;
    $maxrec = 1000; # Suche begrenzen
    my $t = $top->Toplevel(-title => 'Telefonbuch-CD-ROM');
    my($str, $hnr);
    my $e_str = $t->LabEntry(-label => 'Strae',
			     -labelPack => [-side => 'left'],
			     -textvariable => \$str)->pack(-anchor => 'w');
    my $e_hnr = $t->LabEntry(-label => 'Hausnummer',
			     -labelPack => [-side => 'left'],
			     -width => 6,
			     -textvariable => \$hnr)->pack(-anchor => 'w');
    my $l_found;
    my $b_search =
      $t->Button(-text => 'Suchen',
		 -command => sub {
		     $t->Busy;
		     eval {
			 my $res = search_street_hnr($str, $hnr);
			 if (!defined $res || ref $res ne 'ARRAY') {
			     $l_found->configure(-text => 'Nichts gefunden');
			     return;
			 }
			 my($str0, $hnr0, $plz, $ort, $x, $y) = @{ $res };
			 if ($VERBOSE) {
			     print STDERR "Mark $str0 $hnr0 ($x,$y)\n";
			 }
			 if ($mark_sub) {
			     $mark_sub->($x, $y);
			 }
			 $l_found->configure(-text => "$str0 $hnr0 ($plz)");
		     };
		     $t->Unbusy;
		 })->pack;
    $l_found = $t->Label->pack;

    $e_str->focus;
    $e_str->bind('<Return>' => sub { $e_hnr->focus });
    $e_hnr->bind('<Return>' => sub { $b_search->invoke });
    $t->bind('<Escape>' => sub { $t->destroy });
}

sub search_street_hnr {
    my($str, $hnr) = @_;
    my @res = search_string($str, $str_base, 0, [0]);
    if (@res && defined $hnr && $hnr ne '') {
	foreach (@res) {
	    if ($_->[1] eq $hnr) {
		convert($_);
		return $_;
	    }
	}
    }
    if (@res) {
	convert($res[0]);
	return $res[0];
    }
    ();
}

sub convert {
    my $data_ref = shift;
    ($data_ref->[4], $data_ref->[5]) =
      $Karte::GIS::obj->map2standard($data_ref->[4], $data_ref->[5]);
}

sub convert2 {
    my $data_ref = shift;
    ($data_ref->[3], $data_ref->[4]) =
      $Karte::GIS::obj->map2standard($data_ref->[3], $data_ref->[4]);
}

sub tk_tel_dialog {
    my($top, $mark_sub) = @_;
    return if !check_exists($top);
    require Tk::LabEntry;
    $maxrec = 200; # Suche begrenzen
    my $t = $top->Toplevel(-title => 'Telefonbuch-CD-ROM');
    my($nn, $vn);
    my %search_args;
    $search_args{-exact} = 1;
    $ {$search_args{-oldlinear}} = undef;
    my $e_nn = $t->LabEntry(-label => 'Nachname',
			    -labelPack => [-side => 'left'],
			    -textvariable => \$nn)->pack(-anchor => 'w');
    my $e_vn = $t->LabEntry(-label => 'Vorname',
			    -labelPack => [-side => 'left'],
			    -textvariable => \$vn)->pack(-anchor => 'w');
    $t->Checkbutton(-text => 'Exakt',
		    -variable => \$search_args{-exact},
		   )->pack(-anchor => 'w');
    $t->LabEntry(-label => 'Max. Treffer:',
		 -labelPack => [-side => 'left'],
		 -width => 6,
		 -textvariable => \$maxrec)->pack(-anchor => 'w');
    my $l_found;
    my($b_prev, $b_next, $b_weiter);
    my @res;
    my $res_i = 0;

    my $show_sub = sub {
	my(%args) = @_;
	my $nolabel = delete $args{-nolabel};
	my $fast    = delete $args{-fast} || 0;
	if ($res_i > $#res) {
	    warn "Adjusting index from $res_i to $#res";
	    $res_i = $#res;
	}
	if (ref $res[$res_i] ne 'ARRAY') {
	    warn "Not an array reference (index $res_i)";
	    return;
	}
	my(@data) = @{ $res[$res_i] };
	if ($VERBOSE && !$fast) {
	    print STDERR join("||", @data), "\n";
	}
	if (!$fast) {
	    $b_prev->configure(-state => 
			       ($res_i == 0 ? 'disabled' : 'normal'));
	    $b_next->configure(-state => 
			       ($res_i == $#res ? 'disabled' : 'normal'));
	}
	if (defined $mark_sub && defined $data[3] && $data[3] ne '') {
	    convert2(\@data);
	    my($x, $y) = ($data[3], $data[4]);
	    if ($VERBOSE && !$fast) {
		print STDERR "Mark $x/$y\n";
	    }
	    $args{-addtag} = ["$data[12] $data[21], $data[20]",
			      "$data[13] $data[14]"];
	    eval {
		$mark_sub->($x, $y, %args);
	    };
	}
	if (!$nolabel) {
	    my $text = $data[18];
	    $text =~ s/<tr>/\n/g;
	    $text =~ s/<[^>]+>/ /g;
	    $l_found->configure(-text => $text);
	}
    };

    my $b_f = $t->Frame->pack(-anchor => 'w');
    my $b_search =
      $b_f->Button(-text => 'Suchen',
		   -command => sub {
		       $t->Busy;
		       eval {
			   $ {$search_args{-oldlinear}} = undef;
			   @res = search_name($nn, $vn, %search_args);
			   if (!@res || ref $res[0] ne 'ARRAY') {
			       $l_found->configure(-text => 'Nichts gefunden');
			       $b_weiter->configure(-state => 'disabled');
			       return;
			   }
			   $b_weiter->configure
			     (-state => (defined $ {$search_args{-oldlinear}}
					 ? 'normal'
					 : 'disabled'));
			   $res_i = 0;
			   $show_sub->();
		       };
		       $t->Unbusy;
		   })->pack(-side => 'left');
    $b_weiter =
      $b_f->Button(-text => 'Weiter',
		   -command => sub {
		       $t->Busy;
		       eval {
			   @res = search_name($nn, $vn, %search_args);
			   if (!@res || ref $res[0] ne 'ARRAY') {
			       $l_found->configure(-text => 'Nichts gefunden');
			       $b_weiter->configure(-state => 'disabled');
			       return;
			   }
			   $b_weiter->configure
			     (-state => (defined $ {$search_args{-oldlinear}}
					 ? 'normal'
					 : 'disabled'));
			   $res_i = 0;
			   $show_sub->();
		       };
		       $t->Unbusy;
		   })->pack(-side => 'left');
    eval { require Tk::FireButton;
	   Tk::FireButton->VERSION(0.04);
       };
    my $firebutton = (!$@ ? 'FireButton' : 'Button');
    $b_f->Button(-text => '|<',
		 -command => sub {
		     return if !@res;
		     $res_i = 0;
		     $show_sub->();
		 },
		 -padx => 0,
		 -pady => 0,
		)->pack(-side => 'left');
    $b_prev = $b_f->$firebutton(-text => '<' . '<',
				-command => sub {
				    return if !@res;
				    return if $res_i-1 < 0;
				    $res_i--;
				    $show_sub->();
				},
				-padx => 0,
				-pady => 0,
			       )->pack(-side => 'left');
    $b_next = $b_f->$firebutton(-text => '>>',
				-command => sub {
				    return if !@res;
				    return if $res_i+1 > $#res;
				    $res_i++;
				    $show_sub->();
				},
				-padx => 0,
				-pady => 0,
			       )->pack(-side => 'left');
    $b_f->Button(-text => '>|',
		 -command => sub {
		     return if !@res;
		     $res_i = $#res;
		     $show_sub->();
		 },
		 -padx => 0,
		 -pady => 0,
		)->pack(-side => 'left');
    $b_f->Button(-text => 'Alle',
		 -command => sub {
		     return if !@res;
		     $t->Busy;
		     eval {
			 for($res_i = 0; $res_i <= $#res; $res_i++) {
			     $show_sub->(-dont_delete_old => 1,
					 -dont_center => 1,
					 -nolabel => 1,
					 -fast => 1);
			 }
		     };
		     $t->Unbusy;
		 })->pack(-side => 'left');
    if (is_in_path("dial")) {
	$b_f->Button(-text => 'Whlen',
		     -command => sub {
			 my $telnr = $res[$res_i]->[20];
			 if (defined $telnr && $telnr ne '') {
			     print STDERR "Whle $telnr...\n";
			     system("dial $telnr");
			 }
		     })->pack(-side => 'left');
    }

    $l_found = $t->Label(-wraplength => 320,
			 -width => 40,
			 -justify => 'left')->pack;

    $e_nn->focus;
    $e_nn->bind('<Return>' => sub { $e_vn->focus });
    $e_vn->bind('<Return>' => sub { $b_search->invoke });
    $t->bind('<Escape>' => sub { $t->destroy });

}

sub search_name {
    my($nn, $vn, %search_args) = @_;
    my @res = search_string($nn, $tel_base,
			    [26, 12], [12, 13, 18, 21], \&tocmp2,
			    %search_args);
    if (@res && defined $vn && $vn ne '') {
	my @newres;
	$vn = tocmp2($vn);
	foreach (@res) {
	    if (tocmp2($_->[21]) eq $vn) {
		push @newres, $_;
	    }
	}
	return @newres;
    }
    if (@res) {
	return @res;
    }
    undef;
}

sub search_string {
    my($search, $base, $key, $convertref, $tocmp, %search_args) = @_;

    $tocmp = \&tocmp unless defined $tocmp;

    my $fakesearch = $search = $tocmp->($search);
    my $ch = substr($fakesearch, length($fakesearch)-1, 1);
    $ch = chr(ord($ch)-1);
    $fakesearch =~ s/.$/$ch/;
    $fakesearch .= ("\xff" x 10);
    print STDERR "FAKESEARCH: $fakesearch\n" if $VERBOSE;

    my @res;

    my $dir = Telefonbuch::exists();
    if (!defined $dir) {
	die "Kein Verzeichnis gefunden!";
    }

    my $datafile    = "$dir/${base}d00";
    my $indexfile   = "$dir/${base}i00";

    open(I, $indexfile)   or die "$indexfile: $!";
    open(D, $datafile)    or die "$datafile: $!";

    my $begin = 8;
    my $middle;
    seek(I, 0, 2) or die "Can't seek: $!";
    my $end = tell I;
    my $last_middle = -1;

    my $do_linear = 0;

    if (exists $search_args{-oldlinear} &&
	defined $ {$search_args{-oldlinear}}) {
	print STDERR "Continue linear search\n" if $VERBOSE;
	$do_linear = 1;
	$middle = $ {$search_args{-oldlinear}};
    }

    while (1) {
	if (!$do_linear) {
	    $middle = int(($end-$begin)/4/2)*4 + $begin;
	
	    if ($middle == $last_middle) {
		$do_linear = 1;
		print STDERR "Switch to linear search\n" if $VERBOSE;
		$middle += 4;
	    }
	} else {
	    $middle += 4;
	}

	my $buf;

	seek(I, $middle, 0) or die "Can't seek: $!";
	defined read(I, $buf, 4) or die "Can't read: $!";
	my $seek = unpack("V", $buf);
	seek(D, $seek, 0) or die "Can't seek: $!";
	defined read(D, $buf, 4) or die "Can't read: $!";
	my($ds, $extra_offset) = unpack("v2", $buf); # Anzahl der Datenstze
	my $curr_seek = $seek + 4;
	my(@ds, @ds2, @ds3);
	for(my $i = 0; $i < $ds; $i++) {
	    if (int($curr_seek / pagesize) != int(($curr_seek+3) / pagesize)) {
		# next file page
		$curr_seek = (int($curr_seek / pagesize)+1)*pagesize;
		seek(D, $curr_seek, 0) or die "Can't seek: $!";
	    }
	    defined read(D, $buf, 4) or die "Can't read: $!";
	    my($index, $offset) = unpack("v2", $buf);
# 	    if ($index < 0 || $index > $ds) {
# 		print STDERR "*** ERROR\n";
# 		printf STDERR "Invalid index: $index at 0x%x\n", $seek;
# 		exit 1;
# 	    }
#           $ds[$index] = $offset; # XXX
	    $ds[$i] = $offset;
	    $ds2[$i] = $index;
	    $ds3[$index] = $i;
#warn "*** $ds3[$index]\n";
	    $curr_seek += 4;
	}
	$ds[$ds] = 0; # Ende markieren
	my(@data, @data2);
	for(my $i = 0; $i < $ds; $i++) {
	    seek(D, $seek-$ds[$i]-$extra_offset, 0) or die "Can't seek: $!";
	    defined read(D, $buf, $ds[$i]-$ds[$i+1]) or die "Can't read: $!";
	    push @data, $buf;
	    $data2[$ds2[$i]] = $buf;
	}
#print STDERR ">", join("|| ", map { if (ref $_ eq 'ARRAY') { join("|", @$_)} else { $_ } } @data2), "\n" if $VERBOSE;

	my $k;
	if (ref $key eq 'ARRAY') {
	    foreach my $kk (@$key) {
		if (defined $ds3[$kk] && $data[$ds3[$kk]] ne '') {
		    $k = $ds3[$kk];
		    last;
		}
	    }
	} else {
	    $k = $ds3[$key];
	}
	if (!defined $k) {
	    die "No valid key?! (seek $seek)";
	}
	$data[$k] = cp850__latin1($data[$k]);
	my $cmpdata = $tocmp->($data[$k]);
	print STDERR ">", join(", ", @data), "\n" if $VERBOSE;

	if ($do_linear) {
	    if (defined $convertref) {
		foreach my $i (@$convertref) {
		    next if $i == $k; # bereits convertiert
		    $data2[$i] = cp850__latin1($data2[$i]);
		}
	    }
#warn "$cmpdata <=> $search exact:$search_args{-exact}\n";
	    if ((!$search_args{-exact} && $cmpdata =~ /^\Q$search\E/i) ||
		($search_args{-exact}  && $cmpdata =~ /^\Q$search\E$/i)) {
		push @res, [@data2];
		if (defined $maxrec && @res > $maxrec) {
		    print STDERR "ber $maxrec Datenstze gefunden!\n"
		      if $VERBOSE;
		    if (exists $search_args{'-oldlinear'}) {
			$ {$search_args{'-oldlinear'}} = $middle;
		    }
		    last;
		}
#warn "pushing $data2[0]";
	    } else {
#warn "last";
		if (exists $search_args{'-oldlinear'}) {
		    $ {$search_args{'-oldlinear'}} = undef;
		}
		last if ($cmpdata !~ /^$search/i);
	    }
	} else {
	    my $cmp = ($cmpdata cmp $fakesearch);
#warn "$cmpdata $cmp $search";
	    if ($cmp < 0) {
		$begin = $middle;
	    } elsif ($cmp > 0) {
		$end = $middle;
	    } else {
		$do_linear = 1;
		next;
	    }
	    $last_middle = $middle;
	}
    }

    close D;
    close I;

    return @res;
}

sub cp850__latin1 {
    $_ = $_[0];
    if (defined $_) {
	tr/\200\201\202\203\204\205\206\207\210\211\212\213\214\215\216\217\220\221\222\223\224\225\226\227\230\231\232\233\234\235\236\237\240\241\242\243\244\245\246\247\250\251\252\253\254\255\256\257\260\261\262\263\264\265\266\267\270\271\272\273\274\275\276\277\300\301\302\303\304\305\306\307\310\311\312\313\314\315\316\317\320\321\322\323\324\325\326\327\330\331\332\333\334\335\336\337\340\341\342\343\344\345\346\347\350\351\352\353\354\355\356\357\360\361\362\363\364\365\366\367\370\371\372\373\374\375\376\377/\307\374\351\342\344\340\345\347\352\353\350\357\356\354\304\305\311\346\306\364\366\362\373\371\377\326\334\370\243\330\327F\341\355\363\372\361\321\252\272\277\256\254\275\274\241\253\273\.\:\?\|\+\301\302\300\251\+\|\+\+\242\245\+\+\+\+\+\-\+\343\303\+\+\+\+\+\=\+\244\360\320\312\313\310i\315\316\317\+\+FL\246\314T\323\337\324\322\365\325\265\336\376\332\333\331\375\335\-\264\255\261\=\276\266\247\367\270\260\250\'\271\263\262f\240/;
    }
    $_;
}

sub tocmp {
    my $a = shift;
    $a =~ tr//aouAOUs/;
    lc($a);
}

sub tocmp2 {
    my $a = shift;
    if (defined $a) {
	$a =~ s//ae/g;
	$a =~ s//oe/g;
	$a =~ s//ue/g;
	$a =~ s//Ae/g;
	$a =~ s//Oe/g;
	$a =~ s//Ue/g;
	$a =~ s//ss/g;
	return lc($a);
    }
}

# sub ftfile {
#my $ftindexfile = "/cdrom/database/${base}x00";
#     open(X, $ftindexfile) or die "$ftindexfile: $!";
# #for (my $begin = 0x0000; $begin <= 0x1c000*20; $begin += 0xe000) {
#     for my $begin (0x00000, 0x0e000, 0x10000, 0x1c000, 0x28000,
# 		   0x36000, 0x42000, 0x4e000, 0x5a000, 0x66000,
# 		   0x72000, 0x80000, 0x8a000, 0xa0000,
# 		  ) {
# 	#warn "-----------------------$begin";
# 	seek(X, $begin, 0);
# 	my $buf;
# 	read(X, $buf, 0x10);
# 	my $type = unpack("c", substr($buf, 0, 1));
# 	my($page);
# 	read(X, $page, 0x2000-0x10);
# 	my $p = 0;
# 	my @pos;
# 	while(1) {
# 	    my($s) = unpack("s", substr($page, $p, 2));
# 	    last if $s == 0;
# 	    push @pos, $s;
# 	    #printf STDERR "%X\n", $s;
# 	    $p+=2;
# 	}
	
# 	#    if ($begin % 0x1c000 == 0) {
# 	if ($type == 2) {
# 	    my $last_string = '';
# 	    for(my $i=0; $i<=$#pos; $i++) {
# 		my $pos  = $pos[$i];
# 		my $next = ($i > 0 ? $pos[$i-1] : 0x2000-0x10);
# 		my($data, $string) = (substr($page, $pos, 8),
# 				      substr($page, $pos+8, $next-$pos-8));
# 		my $last_string_count = unpack("c", substr($data, 7, 1));
# 		$last_string = substr($last_string, 0, $last_string_count) . $string;
# 		printf "%s %s\n", unpack("H*", $data), $last_string;
# 	    }
# 	} elsif ($type == 1) {
# 	    for(my $i=0; $i<=$#pos; $i++) {
# 		my $pos  = $pos[$i];
# 		my $next = ($i > 0 ? $pos[$i-1] : 0x2000-0x10);
# 		my($data, $string) = (substr($page, $pos, 7),
# 				      substr($page, $pos+7, $next-$pos-7));
# 		printf "%s %s\n", unpack("H*", $data), $string;
# 	    }
# 	} else {
# 	    warn "*** Unknown type $type";
# 	}
#     }
    
#     close X;
# }

1;

__END__
