#ifdef __cplusplus
extern "C" {
#endif
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#ifdef __cplusplus
}
#endif

/* berechnet StrassenNetz mit KoordXY-Komponente, wenn definiert */
#undef WITH_KOORDXY

/* XXX andere Compiler ? */
#ifndef __inline__
#ifndef __GNUC__
#define __inline__
#endif /* __GNUC__ */
#endif /* __inline__ */

#include <math.h>
#include <stdlib.h>

typedef HV* StrassenNetz;

double canvas_scale = 1;

__inline__ static void
to_koord1(s, x, y)
char *s;
SV **x, **y;
{
    char *p;

    p = s;
    while(*p != ',') p++;
    p++;

    *x = newSViv(atoi(s));
    *y = newSViv(atoi(p));
}

static AV*
to_koord(raw_coords)
AV* raw_coords;
{
    int i = 0;
    int len = av_len(raw_coords);
    AV *res = newAV();

    for(; i<=len; i++) {
      SV **tmp;
      SV *x, *y;
      AV *elem;
      char *s;

      tmp = av_fetch(raw_coords, i, 0);
      s = SvPV(*tmp, na);

      to_koord1(s, &x, &y);
      elem = newAV();
      av_push(elem, x);
      av_push(elem, y);
      av_push(res, newRV_noinc((SV*)elem));
    }
    return res;
}

static int
strecke(kreuz_coord, i)
AV* kreuz_coord;
int i;
{
    double a1, a2;
    SV *tmp1, *tmp2, **tmp3;
    int x1,y1,x2,y2;

    tmp1 = SvRV(*av_fetch(kreuz_coord, i, 0));
    tmp2 = SvRV(*av_fetch(kreuz_coord, i+1, 0));

    a1 = SvIV(*(av_fetch((AV*)tmp1, 0, 0))) -
         SvIV(*(av_fetch((AV*)tmp2, 0, 0)));
    a2 = SvIV(*(av_fetch((AV*)tmp1, 1, 0))) -
         SvIV(*(av_fetch((AV*)tmp2, 1, 0)));
    return sqrt(a1*a1 + a2*a2);
}

MODULE = BBBikeXS		PACKAGE = main

PROTOTYPES: DISABLE


void
set_canvas_scale_XS(scale)
	double scale;

	PPCODE:
	canvas_scale = scale;


AV*
transpose_ls_XS(x, y)
	int x;
	int y;

	PPCODE:
	EXTEND(sp, 2);
	PUSHs(sv_2mortal(newSViv((-200.0+x/25.0)*canvas_scale+0.5)));
	PUSHs(sv_2mortal(newSViv((600.0-y/25.0)*canvas_scale+0.5)));



MODULE = BBBikeXS		PACKAGE = Strassen

PROTOTYPES: DISABLE


AV*
to_koord1_XS(s)
	char *s;

	PREINIT:
	AV *elem;
	SV *x, *y;

	PPCODE:
	to_koord1(s, &x, &y);
	EXTEND(sp, 2);
	PUSHs(sv_2mortal(x));
	PUSHs(sv_2mortal(y));


MODULE = BBBikeXS		PACKAGE = StrassenNetz

PROTOTYPES: DISABLE

void
make_net_XS(self, ...)
	StrassenNetz self;

	PREINIT:
	HV *net, *net2name;
#ifdef WITH_KOORDXY
	HV *koordxy;
#endif
	SV *strassen, *ret;
	SV** tmp;
	int dbg_i = 0;
	char *k0_s, *k1_s;
	STRLEN k0_slen, k1_slen;

	CODE:
	net      = newHV();
	net2name = newHV();
#ifdef WITH_KOORDXY
	koordxy  = newHV();
#endif
	hv_store(self, "Net",      strlen("Net"),
		       newRV_noinc((SV*) net), 0);
	hv_store(self, "Net2Name", strlen("Net2Name"),
		       newRV_noinc((SV*) net2name), 0);
#ifdef WITH_KOORDXY
	hv_store(self, "KoordXY",  strlen("KoordXY"),
		       newRV_noinc((SV*) koordxy), 0);
#endif
	tmp = hv_fetch(self, "Strassen", strlen("Strassen"), 0);
	if (tmp == NULL)
	    croak("Missing $self->{Strassen}.\n");
	if (!SvROK(*tmp) || SvTYPE(SvRV(*tmp)) != SVt_PVHV)
	    croak("$self->{Strassen} is not a valid reference.\n");
	strassen = *tmp;

	PUSHMARK(sp);
	XPUSHs(strassen);
	PUTBACK;
	perl_call_method("init", G_DISCARD|G_VOID);

	while(1) {
	  AV *kreuzungen, *kreuz_coord;
	  int i, kreuzungen_len;
	
	  ENTER;
	  SAVETMPS;
	  PUSHMARK(sp);
	  XPUSHs(strassen);
	  PUTBACK;
	  perl_call_method("next", G_SCALAR);
	  SPAGAIN;
	  ret = newSVsv(POPs);
	  PUTBACK;
	  FREETMPS;
	  LEAVE;

	  tmp = av_fetch((AV*)SvRV(ret), 1, 0);
	  
	  if (tmp == NULL) break; /* XXX error? */
	  kreuzungen = (AV*)SvRV(*tmp);
	  if (kreuzungen == NULL || av_len(kreuzungen) == -1) break;
	  kreuz_coord = to_koord(kreuzungen);

	  kreuzungen_len = av_len(kreuzungen);
	  for(i = 0; i < kreuzungen_len; i++) {
	    SV **tmp, *k0, *k1;
	    HV *hashtmp;
	    int entf = strecke(kreuz_coord, i);
	    k0 = *av_fetch(kreuzungen, i, 0);
	    k1 = *av_fetch(kreuzungen, i+1, 0);
	    k0_s = SvPV(k0, k0_slen);
	    k1_s = SvPV(k1, k1_slen);

	    tmp = hv_fetch(net, k0_s, k0_slen, 0);
	    if (tmp == NULL) {
	      hashtmp = newHV();
	      hv_store(net, k0_s, k0_slen, newRV_noinc((SV*)hashtmp), 0);
	    } else
	      hashtmp = (HV*)SvRV(*tmp);
	    hv_store(hashtmp, k1_s, k1_slen, newSViv(entf), 0);

	    tmp = hv_fetch(net, k1_s, k1_slen, 0);
	    if (tmp == NULL) {
	      hashtmp = newHV();
	      hv_store(net, k1_s, k1_slen, newRV_noinc((SV*)hashtmp), 0);
	    } else
	      hashtmp = (HV*)SvRV(*tmp);
	    hv_store(hashtmp, k0_s, k0_slen, newSViv(entf), 0);
#ifdef WITH_KOORDXY
	    if (!hv_exists(koordxy, k0_s, k0_slen)) {
	      tmp = av_fetch(kreuz_coord, i, 0);
	      hv_store(koordxy, k0_s, k0_slen, newSVsv(*tmp), 0);
	    }
#endif
	    tmp = hv_fetch(net2name, k0_s, k0_slen, 0);
	    if (tmp == NULL) {
	      hashtmp = newHV();
	      hv_store(net2name, k0_s, k0_slen, newRV_noinc((SV*)hashtmp), 0);
	    } else
	      hashtmp = (HV*)SvRV(*tmp);
	    tmp = hv_fetch((HV*)SvRV(strassen),
			   "Pos", strlen("Pos"), 0);
	    hv_store(hashtmp, k1_s, k1_slen, newSVsv(*tmp), 0);
	  }
#ifdef WITH_KOORDXY
	  /* letztes $i */
	  if (!hv_exists(koordxy, k1_s, k1_slen)) {
	    tmp = av_fetch(kreuz_coord, i, 0);
	    hv_store(koordxy, k1_s, k1_slen, newSVsv(*tmp), 0);
	  }
#endif

	  av_undef(kreuz_coord);
	  SvREFCNT_dec(ret); /* XXX ja? */

	}


